/**
 * Illustrates the impact of coalesced global memory accesses
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 256



/// copies a n x n matrix (in) into out (no coalescing!)
__global__ void copy_bad(float * const in, float *out, int cols, int rows);

/// copies a n x n matrix (in) into out (with coalescing!)
__global__ void copy_good(float * const in, float *out, int cols, int rows);


int main(int argc, char** argv)
{
   int i;
   int rows = 4096;
   int cols = 4096;
   int n = rows * cols;
   if(argc > 1)
      n = atoi(argv[1]);

   if(rows%BLOCKSIZE != 0){
      printf("#Rows must be divisible by %d\n",BLOCKSIZE);
      exit(-1);
   }

   int size = sizeof(float) * n;
   printf("Allocates %f MB\n", 2.0f * ((float)size)/1024/1024);


   //allocate host memory
   float *A_h, *B_h;
   A_h = (float*) malloc(size);
   B_h = (float*) malloc(size);


   //init x and y with random numbers in [0,1]
   for(i=0;i < n; ++i){
      A_h[i] = ((float)rand())/RAND_MAX;
   }


   //allocate device memory
   float *A_d, *B_d;
   HANDLE_ERROR( cudaMalloc(&A_d, size) );
   HANDLE_ERROR( cudaMalloc(&B_d, size) );


   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(A_d, A_h, size, cudaMemcpyHostToDevice) ); 



   float time;
   cudaEvent_t start, stop; //used for timings
   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); //insert the start event into stream 0

   //launch 1 thread block per col
#ifdef BAD
   copy_bad<<<cols, BLOCKSIZE>>>(A_d, B_d, cols, rows); //this kernel is implicitly inserted into stream 0
#endif

   //launch 1 thread block per row
#ifdef GOOD
   copy_good<<<rows, BLOCKSIZE>>>(A_d, B_d, cols, rows); //this kernel is implicitly inserted into stream 0
#endif
   HANDLE_ERROR(cudaGetLastError());

   HANDLE_ERROR( cudaEventRecord(stop, 0) ); //insert stop event into stream 0
   HANDLE_ERROR( cudaEventSynchronize(stop) ); //wait till the stop event has occured
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); //take time difference between start and stop event
   printf("Time:  %3.1f ms \n", time);


   //copy data from device to host
   HANDLE_ERROR( cudaMemcpy(B_h, B_d, size, cudaMemcpyDeviceToHost) ); 

   //free memory
   free(A_h); 
   cudaFree(A_d); cudaFree(B_d);

   return 0;
}



__global__ void copy_bad(float * const a, float *b, int cols, int rows)
{

   int idy = threadIdx.x; 
   int idx = blockIdx.x;

   for(; idy < rows; idy += blockDim.x)
      b[idy * cols + idx] = a[idy * cols + idx];

}



















__global__ void copy_good(float * const a, float *b, int cols, int rows)
{

   int idx = threadIdx.x; 
   int idy = blockIdx.x;

   for(; idx < cols; idx += blockDim.x)
      b[idy * cols + idx] = a[idy * cols + idx];

}









